/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2002, 2003 Tony Finch <dot@dotat.at>.                   *
 *   All rights reserved.                                                  *
 *   Copyright (C) 1985, 1993 The Regents of the University of California. *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2013 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed by Mike Kinghan, imk@burroingroingjoing.com,              *
 *   derived from the code of Tony Finch                                   *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "chew.h"
#include "options.h"
#include "line_edit.h"
#include "diagnostic.h"
#include "io.h"
#include "symbol.h"
#include <cstring>

using namespace std;

/*!
    \file chew.cpp
    This file implements the class `chewer_base` and
	the specializations of template class `chewer`
*/

chewer_base::comment_state
chewer_base::_comment_state_(NO_COMMENT);
bool chewer_base::_in_double_quote_(false);
chewer_base::line_state chewer_base::_line_state_(LS_NEUTER);
unsigned chewer_base::_last_quote_start_line_(0);
unsigned chewer_base::_last_comment_start_line_(0);
bool chewer_base::_stop_at_quote_(false);
bool chewer_base::_must_balance_quotes_(true);

void chewer_base::set_last_comment_start_line()
{
	_last_comment_start_line_ = line_edit::line_num();
}

template<>
chewer<continuation>::chewer(char const *cp)
{
	size_t nl_len;
	for (	; cp[0] == '\\' && (nl_len = eol(cp + 1)); cp += nl_len + 1) {
		if (cp[nl_len + 1] == 0) {
			line_despatch::extend();
			cp = io::read_more(cp);
		}
	}
	_end = cp;
}

template<>
chewer<string>::chewer(char const *cp)
{
	char const * save = cp;
	if (*cp == '\"') {
		for ( ++cp ; *cp && *cp != '\"' && !eol(cp);
		      cp = chewer<continuation>(++cp)) {}
		if (!*cp) {
			warning_unclosed_quote() <<
			                         "Unclosed quotation in context \"" << save << '\"'
			                         << emit();
		}
		++cp;
	} else for (	; isgraph(*cp); cp = chewer<continuation>(++cp)) {
		size_t read_off = line_edit::read_offset(cp);
		char const * leader = chewer<discardable>(cp);
		cp = line_edit::read_pos(read_off);
		if (leader != cp) {
			break;
		}
	}
	_end = cp;
}

template<>
chewer<unbroken_string>::chewer(char const *cp)
{
	char const * save = cp;
	if (*cp == '\"') {
		for ( ++cp ; *cp && *cp != '\"' && !eol(cp) &&
		      !is_line_continuation(cp); ++cp) {}
		if (!*cp) {
			warning_unclosed_quote() <<
				 "Unclosed quotation in context \"" << save << '\"'
				 << emit();
		} else if (*cp == '\"') {
			++cp;
		}
	} else for (	; isgraph(*cp) && !is_line_continuation(cp); ++cp) {
			size_t read_off = line_edit::read_offset(cp);
			char const * leader = chewer<discardable>(cp);
			cp = line_edit::read_pos(read_off);
			if (leader != cp) {
				break;
			}
		}
	_end = cp;
}

char const * chewer<discardable>::operator()(char const *cp)
{
	char const *chew_start = cp;
	unsigned nl_len;
	/*  Need to track single-quotes to detect quotations of the
		character '"', which must not count as opening or closing
		double-quotes. Is necessary only to recognise single-quotes
		enclosing a single character: if the enclosed charccter is
		escaped, e.g. '\"', then escape-tracking copes.
	*/
	if (options::plaintext()) {
		for (; isspace((unsigned char)*cp); ++cp) {
			if (eol(cp)) {
				_line_state_ = LS_NEUTER;
			}
		}
		return (cp);
	}
	while (*cp != '\0') {
		if (_single_quote_length) {
			++_single_quote_length;
		}
		if (cp[0] == '\\') {
			if (eol(cp + 1)) {
				/* Line continuation "\\n" sequence cannot itself be escaped */
				_escape = true;
			} else if (_single_quote_length) {
				/* Toggle escape state if we are in single quotes */
				_escape = !_escape;
				/* Discount an escape token from the length of a single
					quotation */
				--_single_quote_length;
			} else if (_in_double_quote_) {
				/* Also Toggle escape state if we are in double quotes */
				_escape = !_escape;
			} else if (_line_state_ != LS_CODE ||
			           _comment_state_ == NO_COMMENT) {
				/* This is a plain backslash within directive. Quit */
				break;
			}
			++cp;
			continue;
		}
		if (cp[0] == '"') {
			if (!_escape) {
				if (_stop_at_quote_ &&
				    _comment_state_ == NO_COMMENT) {
					break;
				}
				/* 	Can toggle double-quote state if we are not escaped.
				    This action will be invalid if this double-quote is
				    within single-quotes, but in that case we will
				    detect and rectify the error when we close the
				    single-quote. */
				if ((_in_double_quote_ = !_in_double_quote_) == true) {
					_last_quote_start_line_ = line_edit::line_num();
				}
			}
			++cp;
		} else if (cp[0] == '\'') {
			if (!_escape) {
				if (_stop_at_quote_ &&
				    _comment_state_ == NO_COMMENT) {
					break;
				}
				/* Can only enter a single-quote state is we are not escaped */
				if (_single_quote_length == 0 && !_in_double_quote_) {
					/* Can enter single-quotation if not already in
					single- or double-quotatioon */
					++_single_quote_length;
				} else if (_single_quote_length > 0) {
					/* Leaving single quotation */
					if (_in_double_quote_) {
						/* If we are leaving single-quotation then we
						cannot have been within double-quotation when
						we entered single-quotation because we forbid
						entry to single-quotation within double-quotation.
						Therefore we have entered double-quotation within
						single-quotation. If the last '"'-character
						parsed was escaped we would not	have entered double
						quotation. And since we are	still within
						single-quotation now, we know that the previous
						character parsed was an unescaped '"', which
						we falsely interpreted as the start of
						double-quotation. So we correct that error
						now by exiting double-quotation.*/
						_in_double_quote_ = false;
					}
					_single_quote_length = 0;
				}
				++cp;
			} else if (_in_double_quote_ || _single_quote_length > 0 ||
			           _comment_state_ != NO_COMMENT) {
				/* If we're in double quotation or comments then apostrophe is
					not start of a  character constant */
				++cp;
			} else { /* Apostrophe may be start of a character constant */
				break;
			}
		} else if ((nl_len = eol(cp)) != 0) { /* Newline*/
			if (_escape) {
				/* Line continuation */
				if (cp[nl_len] == 0) {
					line_despatch::extend();
					/* Newline is at actual end of line, so extend line */
					cp = io::read_more(cp);
				}
				if (_single_quote_length) {
					/*  A line continuation doesn't count as a character,
						so discount both a '\' and preceding '\\' from the
						length of the single-quote */
					_single_quote_length -= 2;
				}
			} else {	/* Plain newline*/
				_line_state_ = LS_NEUTER;
				if (_comment_state_ == CXX_COMMENT) {
					/* Newline terminates C++ comment*/
					_comment_state_ = NO_COMMENT;
					_in_double_quote_ = false;
				} else if (_comment_state_ == C_COMMENT) {
					_in_double_quote_ = false;
					if (cp[nl_len] == 0) {
						line_despatch::extend();
						/* Newline is at actual end of line, so extend line */
						cp = io::read_more(cp);
					}
				} else if (_in_double_quote_) {
					_in_double_quote_ = false;
					if (_must_balance_quotes_) {
						/* Dangling quotation is not in comment. Error*/
						error_newline_in_quote() <<
	                         "Newline within quotation" << emit();
					}
				}
			}
			if (*cp == '\r') {
				/* Detected Windows line-end in Unix environment*/
				++cp;
			}
			++cp;
		} else if (cp[0] == ' ' || cp[0] == '\t') { /* Some whitespace*/
			++cp;
		}
		/* Could be at the start or end of a comment*/
		else if (_comment_state_ == NO_COMMENT) {
			/* Not in C or C++ comment*/
			/* Check for start of comment*/
			if (!_in_double_quote_ && !_single_quote_length) {
				/* Don't let comments start within quotation*/
				if (!strncmp(cp, "/\\\n",3)) {
					_comment_state_ = STARTING_COMMENT;
					_last_comment_start_line_ = line_edit::line_num();
					cp += 3;
				}
				if (!strncmp(cp, "/\\\r\n",4)) {
					_comment_state_ = STARTING_COMMENT;
					_last_comment_start_line_ = line_edit::line_num();
					cp += 4;
				} else if (cp[0] == '/' && cp[1] == '*') {
					_comment_state_ = C_COMMENT;
					_last_comment_start_line_ = line_edit::line_num();
					cp += 2;
				} else if (cp[0] == '/' && cp[1] == cp[0]) {
					_comment_state_ = CXX_COMMENT;
					_last_comment_start_line_ = line_edit::line_num();
					cp += 2;
				} else if (_line_state_ == LS_CODE) {
					/* Cool if we have got into code. Carry on */
					++cp;
				} else {
					/* Got some non-chewable char within directive. Quit */
					break;
				}
			} else {	/* We're inside quotation. Truck on*/
				++cp;
			}
		}
		/* We are in a comment state. Check for end of comment*/
		/* No need to test for end of C++ comment 'cos newline case
			covers it*/
		else if (_comment_state_ == C_COMMENT) {
			if (!strncmp(cp, "*\\\n",3)) {
				_comment_state_ = FINISHING_COMMENT;
				cp += 3;
			}
			if (!strncmp(cp, "*\\\r\n",4)) {
				_comment_state_ = FINISHING_COMMENT;
				cp += 4;
			} else if (cp[0] == '*' && cp[1] == '/') {
				_comment_state_ = NO_COMMENT;
				_in_double_quote_ = false;
				cp += 2;
			} else {
				++cp;
			}
		} else if (_comment_state_ == STARTING_COMMENT) {
			if (*cp == '*') {
				_comment_state_ = C_COMMENT;
				++cp;
			} else if (*cp == '/') {
				_comment_state_ = CXX_COMMENT;
				++cp;
			} else {
				_comment_state_ = NO_COMMENT;
				_line_state_ = LS_CODE;
			}
		} else if (_comment_state_ == FINISHING_COMMENT) {
			if (*cp == '/') {
				_comment_state_ = NO_COMMENT;
				++cp;
			} else {
				_comment_state_ = C_COMMENT;
			}
		} else {
			++cp;
		}
		_escape = false;
		if (_single_quote_length > 3) {
			_single_quote_length = 0;
		}
	}
	if (_escape) {
		_escape = false;
		warning_stray_escape() << "Stray '\\' ignored" << emit();
	}
	if (_in_double_quote_) {
		warning_unclosed_quote() << "Unclosed quotation in context \""
		                         << chew_start << '\"' << emit();
	}
	return cp;
}

/* EOF*/
