#ifndef EXPRESSION_PARSER_H
#define EXPRESSION_PARSER_H
/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2002, 2003 Tony Finch <dot@dotat.at>.                   *
 *   All rights reserved.                                                  *
 *   Copyright (C) 1985, 1993 The Regents of the University of California. *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2013 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed by Mike Kinghan, imk@burroingroingjoing.com,              *
 *   derived from the code of Tony Finch                                   *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "platform.h"
#include "evaluation.h"
#include <vector>

/*! \file expression_parser.h
    This defines class `expression_parser`
*/

/*! \brief Encapsulates parsing of preprocessor expressions.

    Class `expression_parser` encapsulates the evaluation of
    all contexts that require resolution with respect
    to the defined and undefined symbols.

*/
struct expression_parser {
	
	//! Parser control flags
	enum control {
		//! Do parse unknown identifiers as new symbols
		no_new_symbols = 1,
		//! Do not diagnose syntax errors
		no_syntax = 2,
		//! Parse as plaintext
		plaintext = no_new_symbols | no_syntax
	};

	/*! \brief Construct to evaluate text from an arbitrary address.

	    \param	ptext	On entry points to the address from which to
	        parse text. On return receives the address of the
	        unparsed remainder of the text.
		\param flags Control flags. An OR-combination of 
			`expression_parser::control` 

		The constructed object will have parsed the text at `*ptext`.
	*/
	expression_parser(char const **ptext, unsigned flags = 0);
	
	/*! \brief Construct to evaluate a string
		\param str A string to be evaluated.
		\param flags Control flags. An OR-combination of 
			`expression_parser::control` 
		
		The constructed object will have parsed and evaluated any
		text contained by `str`.
		
		`str` is pushed onto the `line_edit` stack before evaluation
		and the stack is popped on completion; thus `str` is edited
		out-of-stream.

	*/
	expression_parser(std::string & str, unsigned flags = false);
	
	/*! \brief Get the result of parsing the expression.
	    \return The `evaluation` that results from parsing.
	*/
	evaluation result() const {
		return _eval;
	}

	//! Cast the object to an `evaluation`.
	operator evaluation () const {
		return result();
	}

private:

	//! Structure representing an operator.
	struct operation {
		//! Points to the preprocessor token that denotes the operator.
		const char *str;
		//! Pointer to function to implement the operator.
		evaluation (*fn)(evaluation const &, evaluation const &);
	};

	/*! \brief Structure of the operator precedence table.

	    Expressions involving binary operators are evaluated
	    in a table-driven way by `binary_op()`. When it evaluates a
		subexpression it calls the inner function with its first argument
		pointing to the next element of the table.

	    Innermost expressions are evaluated by `unary_op()`.
	*/
	struct co_precedent_ops {
		//! Pointer to function for evaluating operands of operator.
		evaluation (expression_parser::* inner)(
		    struct co_precedent_ops const * ops,
		    char const **, struct operation const **pbin_op);
		/*! \brief Array of operators of same precedence associated with
		    the evaluator \c inner*/
		struct operation op[5];
	};
	
	//! Get the `no_new_symbols` flag
	bool parse_no_new_symbols() const {
		return _flags & no_new_symbols;
	}
	
	//! Get the `no_syntax` flag.
	bool parse_no_syntax() const {
		return _flags & no_syntax;
	}

	//! Less-than operator.
	static evaluation op_lt(evaluation const & lhs, evaluation const & rhs) {
		return evaluation(lhs.value() < rhs.value());
	}

	//! Greater-than operator.
	static evaluation op_gt(evaluation const & lhs, evaluation const & rhs) {
		return evaluation(lhs.value() > rhs.value());
	}

	//! Less-than-or-equal operator.
	static evaluation op_le(evaluation const & lhs, evaluation const & rhs) {
		return evaluation(lhs.value() <= rhs.value());
	}

	//! Greater-than-or-equal operator.
	static evaluation op_ge(evaluation const & lhs, evaluation const & rhs) {
		return evaluation(lhs.value() >= rhs.value());
	}

	//! Equality operator.
	static evaluation op_eq(evaluation const & lhs, evaluation const & rhs) {
		return evaluation(lhs.value() == rhs.value());
	}

	//! Inequality operator.
	static evaluation op_ne(evaluation const & lhs, evaluation const & rhs) {
		return evaluation(lhs.value() != rhs.value());
	}

	//! Inclusive-or operator.
	static evaluation op_or(evaluation const & lhs, evaluation const & rhs);

	//! And operator.
	static evaluation op_and(evaluation const & lhs, evaluation const & rhs);
	
	//! A fake function representing the '?' "operator".
	static evaluation
	op_ternary_if_dummy(evaluation const & lhs, evaluation const & rhs) {
		return lhs;
	}

	//! A fake function representing the ':' "operator".
	static evaluation op_ternary_either_or_dummy(evaluation const & lhs,
											evaluation const & rhs) {
		return lhs;
	}
	
	//! Bitwise AND operator.
	static evaluation
	op_bit_and(evaluation const & lhs, evaluation const & rhs) {
		return evaluation(lhs.value() & rhs.value());
	}

	//! Bitwise OR operator.
	static evaluation
	op_bit_or(evaluation const & lhs, evaluation const & rhs) {
		return evaluation(lhs.value() | rhs.value());
	}

	//! Bitwise XOR operator.
	static evaluation
	op_bit_xor(evaluation const & lhs, evaluation const & rhs) {
		return evaluation(lhs.value() ^ rhs.value());
	}

	//! Left-shift operator.
	static evaluation
	op_lshift(evaluation const & lhs, evaluation const & rhs) {
		return evaluation(lhs.value() << rhs.value());
	}

	//! Right-shift operator.
	static evaluation
	op_rshift(evaluation const & lhs, evaluation const & rhs) {
		return evaluation(lhs.value() >> rhs.value());
	}

	//! Addition operator.
	static evaluation op_add(evaluation const & lhs, evaluation const & rhs) {
		return evaluation(lhs.value() + rhs.value());
	}

	//! Subtraction operator.
	static evaluation
	op_subtract(evaluation const & lhs, evaluation const & rhs) {
		return evaluation(lhs.value() - rhs.value());
	}

	//! Multiplication operator.
	static evaluation op_mult(evaluation const & lhs, evaluation const & rhs) {
		return evaluation(lhs.value() * rhs.value());
	}

	//! Division operator.
	static evaluation op_divide(evaluation const & lhs, evaluation const & rhs);

	//! Modulus operator.
	static evaluation op_mod(evaluation const & lhs, evaluation const & rhs);
	
	/*! \brief Evaluate arbitrary text.

	    \param	ptext	On entry points to the address from which to
	        parse text. On return receives the address of the
	        unparsed remainder of the text.

		Parse and evaluate any text commencing at `*ptext`.
	*/
	void parse(char const **ptext);

	/*! \brief Evaluator for innermost subexpressions.

	    The function evaluates symbols, numbers and unary operations, including
	 the parenthesis operation.

	 \param	    ops		Pointer to the `co_precedent_ops` to be used for
	                    the evaluation.
	 \param cpp		On entry, a pointer to the start of the text
	                    to be evaluated. Receives the address reached
	                    by evaluation.
	 \param ignored    The parameter is present only for protopype compatibility
	        with `binary_op()` and is unused.

	 \return	An `evaluation` representing
	            the result of evaluation.
	*/
	evaluation
	unary_op(co_precedent_ops const * ops, char const **cpp,
	         operation const **ignored);

	/*! \brief Evaluator for binary operations.

	    The functions evaluates subexpressions
	    that are binary operations.

	    The evaluator does shortcircuit evaluation for && and ||.
	    It also simplifies binary subexpressions that can be solved
	    only on one side.

	 \param	ops		Pointer to the `co_precedent_ops` to be used for
	            the evaluation.
	 \param cpp	On entry, a pointer to the start of the text
	            to be evaluated. Receives the address reached by evaluation.

	 \param pbin_op Either `NULL` or the address of an `operation const *`
	        at which will be stored on return the adress of the binary operation,
	        if any, that is the major operator in the parsed text.

	 \return	An `evaluation` that representing the result of evaluation.
	*/
	evaluation
	binary_op(co_precedent_ops const *ops, char const **cpp,
	          operation const **pbin_op);
			  
	/*! \brief apply a binary operation to arguments
		\param	op	Pointer to the operation to apply
		\param	lhs	The left-hand argument
		\param	rhs	The right-hand argument
		\return The result of applying `op` to `lhs` and `rhs`
	*/
	evaluation apply(const operation * op, evaluation & lhs, evaluation &rhs);


	/*! \brief Say whether a given operator occurs at a given address.
	    \param ptxt	    On entry, pointer to an address in the parsed text. On
			return, if the operator is matched, stores the address not consumed 
			by the match; otherwise the entry value is unchanged.
	    \param op	Pointer to an operator token to match.
	    \return		`tru`e iff the operator token `op` occurs at position 
		`*ptxt`.
	*/
	static bool match_op(char const ** ptxt, char const *op);

	//! The operator precedence table.
	static const co_precedent_ops eval_ops [];

	//! Pointer to the operator precedence table.
	static co_precedent_ops const * ops_tab;

	//! The \c evaluation that results from parsing the expression.
	evaluation _eval;
	//! Are we parsing a macro argument list?
	unsigned _flags;
	//! Stack of ternary conditional results;
	std::vector<evaluation> _ternary_cond_stack;
};

#endif /* EOF*/
