/***************************************************************************
 *   Copyright (C) 2007-2013 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the arg of Symbian Software Ltd. nor the names of its         *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "platform.h"
#include "hash_include.h"
#include "diagnostic.h"
#include "options.h"
#include "io.h"
#include "line_edit.h"
#include "chew.h"
#include <iostream>

/*! \file hash_include.cpp
    This file implements class `hash_include`.
*/

using namespace std;

char const * hash_include::scan_header_name(char const *cp)
{
	char delim;
	if (*cp == '\"') {
		delim = *cp;
	} else if (*cp == '<') {
		delim = '>';
	} else {
		delim = 0;
	}
	if (delim) {
		for (cp = chewer<continuation>(++cp);
		     *cp != delim &&
		     (symbol::is_valid_char(*cp) ||
		      *cp == ' ' ||
		      *cp == '/' ||
#ifdef WINDOWS
		      *cp == '\\' ||
#endif
		      *cp == '.');
		     cp = chewer<continuation>(++cp)) {};
		if (*cp == delim) {
			cp = chewer<continuation>(++cp);
		}
	}
	return cp;
}

bool hash_include::valid() const
{
	string fn = filename();
	char const * start = fn.data();
	char const * end = scan_header_name(start);
	size_t len = fn.length();
	bool valid = len > 2U && len == size_t(end - start);
	if (valid) {
		--end;
		valid = ((*start == '<' && *end == '>') ||
		         (*start == '\"' && *end == '\"'));
	}
	return valid;
}

string hash_include::read(char const *cp, char const **pend)
{
	string arg;
	char const *end = cp;
	if (symbol::is_start_char(*cp)) {
		arg = symbol::read(cp,&cp);
		argument_list macro_args(cp,&end);
		if (!macro_args.well_formed()) {
			error_malformed_macro()
			        <<"Malformed macro argument list for symbol \""
			        << arg << "\"" << emit();
			return string();
		}
		_parts.first = arg;
		_parts.second = macro_args;
		arg += macro_args.str();
	} else {
		end = chewer<string>(cp);
		arg = string(cp,end);

	}
	if (pend) {
		*pend = end;
	}
	return arg;
}

void hash_include::report()
{
	symbol sym = symbolic_interpretation();
	argument_list const & macro_args = _parts.second;
	if (_directive.reported()) {
		if (sym.configured() && !sym.reported()) {
			_directive.set_reported(false);
		}
	}
	if (_directive.reported() && options::list_only_once()) {
		return;
	}
	if ((options::list_local_includes() == options::list_system_includes()) ||
	    (options::list_local_includes() && local_header()) ||
	    (options::list_system_includes() && system_header())) {
		cout << "#include ";
		if (sym.configured()) {
			if (sym.macro_parameters().length() != macro_args.length()) {
				error_mismatched_macro_args()
				        << "Symbol " << sym.id() << " does not take "
				        << macro_args.length() << " macro parameters" << emit();
				return;
			}
			cout << _directive.argument() << ": defined symbol, ";
			evaluation eval = sym.evaluate(macro_args);
			if (eval.circular()) {
				cout << "insoluble, because of circular definitions";
			} else {
				cout << "resolves as >>" << filename() << "<<";
			}
		} else {
			cout << filename();
		}
		if (options::list_location()) {
			cout << ": " << io::in_file_name() << '(' << line_edit::line_num() << ')';
		}
		cout << '\n';
		_directive.set_reported();
	}
}

/* EOF */
