#ifndef HASH_INCLUDE_H
#define HASH_INCLUDE_H
/***************************************************************************
 *   Copyright (C) 2007-2013 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "platform.h"
#include "symbol.h"
#include "directive.h"

/*! \file hash_include.h
	This file defines class `hash_include`
*/

/*! \brief Class \c hash_include encapsulates an `#include` directive.
*/
struct hash_include {
	/*! \brief Type of container for the parts of a macro-invocation

	    Type of container for the parts of a macro-invocation
	    occurring as the body of the `#include` directive
	*/
	typedef std::pair<std::string,argument_list> symbolic_breakdown;

#if CXX11_HAVE_DECL_DELETE
	//! There is no default constructor
	hash_include() = delete;
	//! Copying a \c hash_include is not allowed
	hash_include(hash_include const & src) = delete;
	//! Assigning a \c hash_include is not allowed
	hash_include & operator=(hash_include const &) = delete;
#else
	hash_include();
	hash_include(hash_include const & src);
	hash_include & operator=(hash_include const &);
#endif

	/*! \brief Explicitly construct from the argument of an
	    `#include`  directive.
	    \param  arg The body of the `#include` directive
	*/
	explicit hash_include(std::string const & arg)
		:  _parts(std::string(),argument_list()),_directive(arg) {}

	/*! \brief Explicitly construct from the argument of an
	    `#include`  directive.
	    \param  cp Pointer to the body of the `#include` directive.
	    \param pend An optional pointer at which shall be stored on return
	        the first address beyond the body of the directive.
	*/
	explicit hash_include(char const *cp, char const **pend = NULL)
		:  _parts(std::string(),argument_list()),_directive(read(cp,pend)) {}


	/*! \brief Say whether the body of an `#include` directive is
	    a configured symbol or macro-invocation.

	    \return If the body of the directive is a configured symbol or
	    macro invocation of a configured symbol, a reference to that configured
	    symbol is returned; otherwise reference to the null symbol.
	*/
	symbol symbolic_interpretation() const {
		return symbol::lookup(_parts.first);
	}


	/*! \brief Return the header file expression.

	    \return If the body of the `#include` is a macro-invocation of
	    a configured symbol then result of resolving that invocation is
	    returned. Otherwise the returned string is the same as returned by
	    `argument()`.
	*/
	std::string filename() const {
		symbol const & sym = symbolic_interpretation();
		return  sym.null() ?
		        _directive.argument() : sym.resolve(_parts.second);
	}


	/*! \brief Test whether the `#include` specifies a system header,
	    i.e. `<headername>`

	    \return True iff the `#include` specifies a system header.
	*/
	bool system_header() const {
		return filename()[0] == '<';
	}

	/*! \brief Test whether the `#include` specifies a local header,
	    i.e. "headername".

	    \return True iff the `#include` specifies a local header.

	    `system_header()` and `local_header()` are not mutually
	    exclusive, as a symbolic `#include` argument may fail
	    to resolve as either.
	*/
	bool local_header() const {
		return filename()[0] == '\"';
	}

	//! Say whether the `#include` directive is valid.
	bool valid() const;

	//! Report the `#include` directive.
	virtual void report();

private:

	/*! \brief Read the body of an `#include` directive from a text pointer.

	    \param  cp Pointer to the body of the `#include` directive.
	    \param pend An optional pointer at which shall be stored on return
	        the first address beyond the body of the directive.

	    \return The parsed argument.
	*/
	std::string read(char const *cp, char const **pend = NULL);


	/*! \brief Parse a putative header file name from a text pointer.
	    \param  cp The start of the putative header name.
	    \return The first character scanned that cannot belong to
	        a system or local header name.
	*/
	static char const * scan_header_name(char const *cp);

	/*! \brief Breakdown of the directive argument into the components of
	    a macro invocation.

	    Breakdown of the directive argument into the components of
	    a macro invocation. - identifier and argument list - if that
	    is what it is. If the directive argument does not have this
	    composition then this structure consists of an empty string and
	    empty argument list.
	*/
	symbolic_breakdown _parts;
	//! The member handles reporting the `#include` directive.
	directive<HASH_INCLUDE> _directive;

};

#endif /* EOF*/
