/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2002, 2003 Tony Finch <dot@dotat.at>.                   *
 *   All rights reserved.                                                  *
 *   Copyright (C) 1985, 1993 The Regents of the University of California. *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2013 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed by Mike Kinghan, imk@burroingroingjoing.com,              *
 *   derived from the code of Tony Finch                                   *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "line_despatch.h"
#include "diagnostic.h"
#include "options.h"
#include "line_edit.h"
#include "io.h"
#include "chew.h"
#include "contradiction.h"
#include "if_control.h"
#include "canonical_string.h"
#include "directive.h"
#include <iostream>
#include <cassert>
#include <cstring>

/*!
    \file line_despatch.cpp
    This file implements class `line_despatch`*/
using namespace std;

unsigned line_despatch::_lines_suppressed_(0);
unsigned line_despatch::_lines_changed_(0);
size_t line_despatch::_drop_run_length_(0);
size_t line_despatch::_indent_(0);
unsigned line_despatch::_extension_lines_(0);

void line_despatch::printline_fast()
{
	*io::output() << line_edit::line_start();
}

void line_despatch::printline_cut()
{
	size_t i = 0;
	for (	;i < _indent_; ++i) {
		*io::output() << line_edit::cur_line_const()[i];
	}
	string output =
	    diagnostic_base::citable_string(line_edit::cur_line_const().c_str() + i);
	_lines_changed_ += _extension_lines_ + 1;
	*io::output() << output << '\n';
	if (options::get_discard_policy() == DISCARD_BLANK) {
		for (	; _extension_lines_; --_extension_lines_) {
			*io::output() << '\n';
		}
	} else if (options::get_discard_policy() == DISCARD_COMMENT) {
		for (	; _extension_lines_; --_extension_lines_) {
			*io::output() << "//coan <\n";
		}
	}
}

void line_despatch::printline()
{
	if (!line_edit::ops_cut()) {
		printline_fast();
	} else {
		printline_cut();
	}
}

void line_despatch::printline_commented_out()
{
	if (!line_edit::ops_cut()) {
		++_lines_changed_;
		if (!_extension_lines_) {
			*io::output() << "//coan < ";
			printline_fast();
		} else {
			_lines_changed_ += _extension_lines_;
			char const *linestart = line_edit::line_start();
			char const *newline;
			for (   ; (newline = strchr(linestart,'\n')) != NULL; ) {
				*io::output() << "//coan < ";
				io::output()->write(linestart,newline - linestart);
				*io::output() << '\n';
				linestart = newline + 1;
			}
		}
	} else {
		*io::output() << "//coan < ";
		printline_cut();
	}
}

void line_despatch::flushline(bool keep)
{
	if (keep ^ options::complement()) {
		printline();
	} else {
		if (options::get_discard_policy() == DISCARD_BLANK) {
			_lines_changed_ += _extension_lines_ + 1;
			*io::output() << '\n';
			for (	; _extension_lines_; --_extension_lines_) {
				*io::output() << '\n';
			}
		} else if (options::get_discard_policy() == DISCARD_DROP) {
			_lines_suppressed_ += _extension_lines_ + 1;
		} else {
			printline_commented_out();
		}
	}
}

void line_despatch::print()
{
	command_code cmd = options::get_command();
	if (cmd == CMD_SOURCE || cmd == CMD_SPIN) {
		if (options::line_directives()) {
			if (_drop_run_length_) {
				int line_num = line_edit::line_num();
				if (_extension_lines_) {
					--line_num;
				}
				*io::output() << "#line " << line_num << '\n';
				--_lines_suppressed_;
				++_lines_changed_;
			}
			_drop_run_length_ = 0;
		}
		flushline(true);
	}
}

void line_despatch::drop()
{
	if (options::get_command() == CMD_SOURCE
		|| options::get_command() == CMD_SPIN) {
		flushline(false);
		if (options::line_directives()) {
			++_drop_run_length_;
		}
	}
}

void line_despatch::substitute(string const & replacement)
{
	if (options::get_command() == CMD_SOURCE
		|| options::get_command() == CMD_SPIN) {
		*io::output() << replacement;
		--_lines_suppressed_;
		++_lines_changed_;
	}
}

line_type line_despatch::next()
{
	char const *cp;
	line_type retval;
	chewer<discardable> chew;

	if (!io::get_line()) {
		contradiction::flush();
		if_control::transition(LT_EOF);
		return LT_EOF;
	}

	line_edit::ops_cut() = false;
	_indent_ = 0;
	/* Assume no simplification possible*/
	retval = LT_PLAIN;
	cp = chew(line_edit::line_start());
	if (chewer_base::get_line_state() == chewer_base::LS_NEUTER) {
		if (*cp == '#') {
			_indent_ = cp - line_edit::line_start();
			chew.set_line_state(chewer_base::LS_DIRECTIVE);
			cp = chew(cp + 1);
		} else if (*cp != '\0') {
			chew.set_line_state(chewer_base::LS_CODE);
			contradiction::flush();
		}
	}
	if (chewer_base::get_comment_state() ==
	    chewer_base::NO_COMMENT &&
	    chew.get_line_state() == chewer_base::LS_DIRECTIVE) {
		chewer_base::set_stop_at_quote(false);
		chewer_base::set_balancing_quotes(true);
		line_edit::keyword_off() = line_edit::read_offset(cp);
		string keyword(canonical<symbol>(cp,&cp));
		line_edit::keyword_len() = keyword.length();		
		retval = directive_base::eval(keyword,&cp);
	}
	if (retval == LT_DIRECTIVE_KEEP) {
		/* A non-if directive is the same as a non-directive for if-control */
		retval = LT_PLAIN;
	}
	if (chewer_base::get_line_state() == chewer_base::LS_CODE) {
		while (*cp != '\0') {
			cp = chew(cp + 1);
		}
	}
	return (retval);
}

/* EOF*/
