#ifndef LINE_EDIT_H
#define LINE_EDIT_H
/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2013 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "platform.h"
#include <list>
#include <string>
#include <cstddef>
#include <cassert>

/*!
    \file line_edit.h
    This file defines class `line_edit`.
*/


/*! \brief Class `line_edit` class provides editing operations on a
	line (or other string) of text.
*/
struct line_edit {

	/*! \brief Editing codes for deleted characters.

	    Character codes inserted into the
	    line-buffer in place of characters
	    provisionally marked for deletion.
	*/
	enum deletion_codes {
	    //! A character provisionally to be deleted.
	    DELETEABLE_CHAR = 1,
	    //! A left-parethesis provisionally marked for deletion.
	    DELETEABLE_LPAREN,
	    //! A right-parethesis provisionally marked for deletion.
	    DELETEABLE_RPAREN
	};
	
	/*! \brief Replace the keyword of the current directive.

	    Replace the keyword of the current directive with another
	    to rectify the syntax for dropped lines, then output
	    the edited directive.

	    \param	replacement		The keyword with which to replace the current
	                            keyword.
		\param len The length of the replacement
	*/
	static void keyword_edit(const char *replacement, size_t len) {
		context().keyword_edit(replacement,len);
	}
	//! Replace the keyword, truncating the text thereafter.
	static void keyword_edit(const char *replacement) {
		context().keyword_edit(replacement);
	}

	/*! \brief Logically delete redundant parentheses.

	    Logically delete a pair of redundant parentheses from a
	    directive in the line-buffer.

	    \param	lparen_off	    Line buffer offset of the '('
	    \param	rparen_off	    Line buffer offset of the ')'

	    Logical deletion is done by inserting `DELETEABLE_LPAREN`
	    in the line-buffer in place of the '(' and `DELETEABLE_RPAREN`
	    in place of the ')'.

	    If evaluation of the directive results in material simplification
	    (i.e. results in deletions other than deletions of redundant
		parentheses), then the logically deleted parentheses will be physically
		deleted when  evaluation is complete.

	    If evaluation of the directive results in no materal simplification
	    then the logically deleted parentheses will finally be restored
	    by restore_all_paren();

	*/
	static void delete_paren(size_t lparen_off, size_t rparen_off) {
		context().delete_paren(lparen_off,rparen_off);
	}

	/*! \brief Restore a prematurely deleted pair of parentheses in the
	        line buffer.

	    \param  lparen_off The offset of the left-hand parenthesis to be
			restored.
	    \param  rparen_off The offset of the right-hand parenthesis to be
			restored.

	    Recursive descent parsing obliges us to delete what appear to be
	    superfluous parentheses in the recursive context, but as recursion
	    unwinds it can transpire that deleted parentheses ought to have
	    been retained.
	*/
	static void restore_paren(size_t lparen_off, size_t rparen_off) {
		context().restore_paren(lparen_off,rparen_off);
	}

	/*! \brief Restore all logically deleted parentheses to the directive.

	    Restore logically deleted parentheses to the directive in the
	    line buffer	when we cannot materially simplify it.

	    This is done on the principle that if we cannot materially simplify
	    the directive then we will leave it alone.

	*/
	static void restore_all_paren() {
		context().restore_all_paren();
	}

	/*!	\brief Logically delete part of the text.
	    \param	start	Offset to start of text to delete
	    \param	end		Offset terminating the text to delet.

	    The chunk to be deleted is only logically deleted by this function.
	    This is efficient because it avoids shuffling down terminal
	    segments of the buffer in the course of evaluation, when the effect of
	    such editing may be obliterated by subsequent deletions. Physical
		deletion is performed only once at the end of evaluation, by removing
		all logically  deleted portions of the line-buffer.

	    The function is only called to delete a truth-functionally redundant
	    operator and one of its operands.
	*/
	static void cut_text(size_t start, size_t end) {
		context().cut_text(start,end);
	}

	//! Truncate the text.
	static void truncate(size_t off) {
		context().truncate(off);
	}

	/*! \brief Say whether a character in the line buffer is deleteable.
	    \param  ch The character to be tested.
	    \return True if `ch` is an editing code inserted in the
	        line buffer in place of a character that can be deleted.
	*/
	static bool is_deleteable(char ch) {
		return ch >= DELETEABLE_CHAR && ch <= DELETEABLE_RPAREN;
	}
	
	//! Get a reference to the keyword offset in the current directive, if any
	static size_t & keyword_off() {
		return context()._keyword_off;
	}
		
	//! Get a reference to length to the keyword of the current directive, if any
	static size_t & keyword_len() {
		return context()._keyword_len;
	}
	
	//! Get an unedited copy of the current input line.
	static std::string pristine_line();

	//! Get a reference to the current source line.
	static std::string & cur_line() {
		return context()._text;
	}
		
	//! Get a const reference to current line buffer as a string
	static std::string const & cur_line_const() {
		return cur_line();
	}
	
	//! Store a pristine copy of the current input line
	static void save_pristine_line() {
		_pristine_line_ = line_start();
	}
	
	//! Get the start of the current text.
	static char * line_start() {
		return context().line_start();
	}

	//! Get the end of the current text.
	static char const * line_end() {
		return context().line_end();
	}

	/*! \brief Get the offset of an address in the current text.
	    \param  cp  An address in the current text.
	    \return The offset of `cp` in the current text.
	*/
	static ptrdiff_t read_offset(char const *cp) {
		return context().read_offset(cp);
	}
	
	/*! \brief Save a position in the current text

	    \param cp		Address in the current text.

	    The position is saved as an offset into the text
	    so that the correct position can be recovered if the
	    text is moved.
	*/
	static void
	save_read_pos(char const *cp) {
		context().save_read_pos(cp);
	}

	/*! \brief Get the last saved text position.

	    \returns cp	The text position last saved by `save_read_pos()`

	    The returned position equates to the last position saved
	    even if the text has moved in the interim.
	*/
	static char * saved_read_pos() {
		return context().saved_read_pos();
	}
	
	/*! \brief Get the address of an offset into the current text.
	    \param  readoff An offset into the current text.
	    \return The address of `readoff`.
	*/
	static char * read_pos(size_t readoff) {
		return context().read_pos(readoff);
	}
	
	//! Get a reference to the current source line number.
	static int & line_num() {
		return context()._line_num;
	}
	
	//! Get a reference to the ops_cut flag.
	static bool & ops_cut() {
		return context()._ops_cut;
	}
		
	//! Forget all contexts
	static void drop() {
		unsigned lineno = line_num();
		_context_stack_.clear();
		push(std::string(),lineno);
	}

	//! Reinitialize the module
	static void top() {
		_context_stack_.clear();
		push(std::string(),0);
	}
		
	//! Pop the current context, if any
	static void pop() {
		if (stacked()) {
			_context_stack_.pop_back();
		}
	}
	
	//! Push a context on the stack
	static void push(std::string const & text, int lineno = 0) {
#if CXX11_HAVE_EMPLACE_BACK
		_context_stack_.emplace_back(text,lineno);
#else
		_context_stack_.push_back(state(text,lineno));
#endif
	}
	
	//! Say whether the current context is in-stream text
	static bool in_stream() {
		return stacked() == 1;
	}

private:

	//! Type encapsulating line editing context
	struct state
	{
		state()
		: 	_line_num(0),_ops_cut(0),_keyword_off(-1),_keyword_len(-1),
			_saved_pos(-1){}
		
		explicit state(std::string const & text, int lineno = 0)
		: 	_text(text),_line_num(lineno),_ops_cut(0),_keyword_off(-1),
			_keyword_len(-1),_saved_pos(-1){}
		
		//! Get the start of the current text.
		char * line_start() {
			return const_cast<char *>(_text.c_str());
		}
		//! Get the end of the current text.
		char const * line_end() {
			return line_start() + _text.length();
		}
		//! Get the offset of an address in the text.
		ptrdiff_t read_offset(char const *cp) {
			return cp - line_start();
		}
		//! Save a position in the current text
		void save_read_pos(char const *cp) {
			_saved_pos = cp - line_start();
		}
		//! \brief Get the last saved text position.
		char * saved_read_pos() {
			return line_start() + _saved_pos;
		}
		//! Get the address of an offset into the current text.
		char * read_pos(size_t readoff) {
			return line_start() + readoff;
		}
		//!	Logically delete part of the text.
		void cut_text(size_t start, size_t end);
		//! Truncate the text
		void truncate(size_t off) {
			assert(off < _text.length());
			_text.erase(off);
		}
		//! Replace a terminal portion of the text
		void tail_edit(size_t off, const char *what) {
			assert(_text.length() > off);
			_text.erase(off);
			_text.append(what);
		}
		//! Replace the keyword at the end of the text.
		void keyword_edit(const char *replacement);
		//! Replace the keyword in the text.
		void keyword_edit(const char *replacement, size_t len);
		//! Logically delete redundant parentheses.
		void delete_paren(size_t lparen_off, size_t rparen_off);		
		//! Restore a prematurely deleted pair of parentheses in the text
		void restore_paren(size_t lparen_off, size_t rparen_off);
		//! Restore all logically deleted parentheses to the directive.
		void restore_all_paren();
		//! Editable text
		std::string _text;
		//! Input source line number, or 0 for out-of-stream text
		int _line_num;
		//! Have operators been deleted?
		bool _ops_cut;
		//! Offset to #-keyword, if any
		size_t	_keyword_off;
		//! Length of the #-keyword, if any
		size_t _keyword_len;	
		//! Saved cursor offset, or -1 if none.
		size_t _saved_pos;
	};
	
	//! Get the number of stacked contexts.
	static size_t stacked() {
		return _context_stack_.size();
	}
	
	//! Get a reference to the current context
	static state & context() {
		return _context_stack_.back();
	}
			//! Stack of edit contexts.
	static std::list<state> _context_stack_;
	//! Unedited copy of the current input line for use in disgnostics
	static std::string _pristine_line_;
	
};

#endif /* EOF*/
