#ifndef PARAMETER_LIST_H
#define PARAMETER_LIST_H
/***************************************************************************
 *   Copyright (C) 2007-2013 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include <string>
#include <vector>

/*! \file parameter_list.h
    This file defines class `parameter_list` and class `argument_list`.
*/

/*! \brief Class `parameter_list` encapsulates a list of formal or actual
    macro parameters.
*/
struct parameter_list {
	//! Default constructor
	parameter_list()
		: _well_formed(true),_variadic(false) {}

	/*! \brief Explicitly construct a given a text pointer.
	    \param cp   Text pointer at which to start parsing the parameter list.
	    \param  pend    Optional address of a pointer at which shall be stored
	        on return the first address not consumed by parsing.
	*/
	explicit parameter_list(char const *cp, char const **pend = NULL)
	: _well_formed(true),_variadic(false) {
		read(cp,pend);
	}

	//! Destructor
	virtual ~parameter_list() {}

	//! Assign a `parameter_list` from another
	bool operator==(parameter_list const & other) const {
		return _params == other._params;
	}

	//! Say whether the `parameter_list` is well-formed
	bool well_formed() const {
		return _well_formed;
	}
	
	//! Say whether the `parameter_list` is variadic
	bool variadic() const {
		return _variadic;
	}

	//! Get the number of parameters in the `parameter_list`
	size_t length() const {
		return _params.size();
	}

	//! Cast the parameter list to its canonical string representation
	std::string str() const;

	//! Get a const iterator to the beginning of the `parameter_list`
	std::vector<std::string>::const_iterator begin() const {
		return _params.begin();
	}

	//! Get a const iterator to the end of the `parameter_list`
	std::vector<std::string>::const_iterator end() const {
		return _params.end();
	}

	/*! \brief Read the `parameter_list` from a text pointer.

	    \param cp   Text pointer at which to start parsing the parameter list.
	    \param  pend    Optional address of a pointer at which shall be stored
	        on return the first address not consumed by parsing.

	    The `parameter_list` is emptied and replaced by parsing the text
	    at `cp`
	 */
	virtual void read(char const *cp, char const **pend = NULL);

protected:
	//! The list of parameters
	std::vector<std::string> _params;
	//! Is the `parameter_list` well-formed?
	bool _well_formed;
	//! Is the `parameter_list` variadic?
	bool _variadic;
};

/*! \brief Class `argument_list` encapsulates a list of macro arguments, i.e.
    arguments to a macro call.
*/
struct argument_list : parameter_list {
	//! Default constructor
	argument_list() {}

	//! Explicitly construct from a parameter list
	explicit argument_list(parameter_list const & parms)
	: parameter_list(parms){}
	
	/*! \brief Explicitly construct given a text pointer.
	    \param cp   Text pointer at which to start parsing the argument list.
	    \param  pend    Optional address of a pointer at which shall be stored
	        on return the first address not consumed by parsing.
	*/
	explicit argument_list(char const *cp, char const **pend = NULL) {
		read(cp,pend);
	}

	/*! \brief Read the `argument_list` from a text pointer.

	    \param cp   Text pointer at which to start parsing the argument list.
	    \param  pend    Optional address of a pointer at which shall be stored
	        on return the first address not consumed by parsing.

	    The `argument_list` is emptied and replaced by parsing the text
	    at `cp`
	 */
	virtual void read(char const *cp, char const **pend = NULL);

	/*! \brief Substitute this `argument_list` into a macro definition.

	    Substitute this `argument_list` into a the definition of
	    a macro with a given `parameter_list`

	    \param  macro_params The `parameter_list` of the macro to be
	            expanded.
	    \param  definition The definition of the macro to be expanded.
	    \return A string representing the macro expansion that results by
	        substitution of this `argument_list` in the macro definition.
	*/
	std::string
	subst(	parameter_list const & macro_params,
	        std::string const & definition) const;

	/*! \brief Say whether a character can validly appear in a member of
	    an argument list.

	    \param  ch  The character to be tested.
	    \return True iff `ch` can validly appear in a member of an
	        argument list.
	*/
	static bool is_valid_char(char ch) {
		return ch && ch != ',' && ch != '(' && ch != ')';
	}

private:
	/*! \brief Append an argument to the argument list.
	    \param  arg    The putative argument to be appended.
	    \return True if `arg` has a non-0-length, else false,
	        in which case the `argument_list` is unmodified.
	*/
	bool append(std::string const & arg);
};

#endif //EOF
