/***************************************************************************
 *   Copyright (C) 2007-2013 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "symbol.h"
#include "expression_parser.h"
#include "diagnostic.h"
#include "line_edit.h"
#include "io.h"
#include "contradiction.h"
#include "if_control.h"
#include "line_select.h"
#include "canonical_string.h"
#include "chew.h"
#include <iostream>
#include <cstring>

/*! \file symbol.cpp
    This file implements class `symbol`.
*/

using namespace std;

symbol::table_entry symbol::_null_symbol_entry_("",symbol::state(UNCONFIGURED));
map<string,symbol::state> symbol::_sym_tab_(&symbol::_null_symbol_entry_,
        &symbol::_null_symbol_entry_ + 1);
unsigned symbol::state::_snapshot_(0);
set<string> _selected_symbols_set_;

void symbol::state::define(std::string const & defn)
{
	delete _defn;
	_defn = new std::string(defn);
	_reported = false;
	_line = line_edit::line_num();
	_snapshot = _snapshot_++;
}

void symbol::state::undef()
{
	delete _defn, _defn = NULL;
	_macro_params = parameter_list();
	_reported = false;
	_line = line_edit::line_num();
	_snapshot = _snapshot_++;
}

void symbol::state::set_resolved_form(	symbol const & sym,
                                        argument_list const & macro_args) const
{
	if (_resolved == NULL || out_of_date() || _macro_params.length()) {
		_snapshot = _snapshot_;
		delete _resolved;
		if (_macro_params.variadic()) {
			string insoluble = sym.id() + macro_args.str();
			warning_variadic_macros_unsupported() <<
				"Sorry, variadic macro expansion is unsupported. "
				<< "\"" << sym.id() << macro_args.str() << "\"" <<
				" will not be resolved" << emit();
			_resolved = new string(insoluble);
			_eval.set_insoluble();
			return;
		}
		if (_defn) {
			if (*_defn == "") {
				_resolved = new string();
				_eval.set_empty();
				return;
			}
			size_t offset = 0;
			_resolved = new string(sym.macro_expand(macro_args));
			symbol next_sym = symbol::find_any_in(*_resolved,offset);
			for (   ; !next_sym.null();
			        next_sym = symbol::find_any_in(*_resolved,offset)) {
				if (sym.id() == next_sym.id()) { // Circular definition
					_eval.set_circular();
					break;
				}
				char const * cp = _resolved->data() +
				                  offset + next_sym.id().length();
				argument_list macro_args(cp,&cp);
				if (line_select::eligible(CMD_SYMBOLS) && 
							options::explain_resolves() && !sym.deselected()) {
					bool existed = !add_pattern(next_sym.id());
					next_sym.report(macro_args,sym.id());
					if (!existed) {
						remove_pattern(next_sym.id());
					}
				}				
				string next_resolved = next_sym.resolve(macro_args);
				evaluation next_eval = next_sym.evaluate(macro_args);
				if (next_eval.circular()) {
					_eval = next_eval;
					break;
				}
				size_t cut_len = (cp - _resolved->data()) - offset;
				size_t nextoff = offset;
				do {
					_resolved->replace(nextoff,cut_len,next_resolved);
					nextoff = next_sym.find_first_in(*_resolved,nextoff);
				} while(nextoff != string::npos && macro_args.length() == 0);
			}
			if (_eval.circular()) {
				warning_circular_sym_def() << "symbol \"" << sym.id()
					   << "\" has a circular definition" << emit();
			}
			evaluation ev = 
				expression_parser(*_resolved,expression_parser::plaintext);
			_eval.set_value(ev.value());
			if (_eval.resolved()) {
				string numeral(diagnostic_base::citable_string(_eval.value()));
				if (numeral != *_resolved) {
					delete _resolved;
					_resolved = new string(numeral);
				}
			}
			return;
		} else {
			_resolved = new string("0");
			_eval.set_value(0);
		}
	}
}


size_t symbol::find_first_in(std::string name, std::string str, size_t start)
{
	size_t namelen = name.length();
	size_t textlen = str.length();
	for ( start = str.find(name,start); start != string::npos;
	      start = str.find(name,start + namelen)) {
		size_t end = start + namelen;
		if ((start == 0 || !symbol::is_valid_char(str[start - 1])) &&
		    (end == textlen || !symbol::is_valid_char(str[end]))) {
			break;
		}
	}
	return start;
}

symbol symbol::find_any_in(string const & str, size_t & start)
{
	size_t offset = start;
	if (offset < str.length()) {
		string::const_iterator iter = str.begin() + offset;
		string::const_iterator end = str.end();
		while(iter != end) {
			if (is_start_char(*iter)) {
				offset = iter - str.begin();
				string name(1,*iter++);
				for (; iter != end && is_valid_char(*iter); name += *iter++) {}
				symbol hit = lookup(name);
				if (!hit.null()) {
					start = offset;
					return hit;
				}
				continue;
			}
			++iter;
		}
	}
	return symbol();
}

string symbol::read(char const *cp, char const **pend)
{
	string symname = canonical<symbol>(cp,pend);
	if (!symname.length()) {
		error_not_identifier() <<
			   "Identifier needed instead of \"" << cp << '\"' << emit();
	}
	return symname;
}

void symbol::report(
	argument_list const & macro_args,string const & dependent) const
{
	if (dependent.empty() && deselected()) {
		return;
	}
	if (!reported() || !options::list_only_once()) {
		string prefix = id() + macro_args.str();
		if (options::resolve_symbols()) {
			if (options::explain_resolves()) {
				if (defn()) {
					prefix += ": definition >>" + *defn() + "<<";
				} else {
					prefix += ": undef";
				}
				if (!dependent.empty()) {
					prefix += ": via " + dependent;
				}
			}
			if(configured()) {
				if (macro_parameters().variadic()) {
					cout << prefix
						<< ": insoluble, because variadic macro parameters are "
						<< "unsupported";
				} else {
					string resolved = resolve(macro_args);
					evaluation eval = evaluate(macro_args);
					if (eval.circular()) {
						cout << prefix 
						<< ": insoluble, because of circular definitions";
					} else {
						cout << prefix 
						<< ": resolves as >>" << resolved << "<<";
					}
				}
			} else {
				cout << prefix 
				<< ": insoluble, because unconfigured";
			}
		} else {
			cout << prefix;
		}
		if (options::list_location()) {
			cout << ": " << io::in_file_name() 
				<< '(' << line_edit::line_num() << ')';
		}
		cout << '\n';
		get_state()._reported = true;
	}
}

void symbol::digest_global_define(char const *cp)
{
	string definition;
	parameter_list macro_params(cp,&cp);
	if (!macro_params.well_formed()) {
		error_malformed_macro() <<
			"Malformed macro parameter list for symbol \""
			<< id() << "\"" << emit();
		return;
	}
	if (*cp == '=') {
		++cp;
	} else if (*cp) {
		error_garbage_arg()
		        << "Garbage in \"-D" << id() << cp << '\"' << emit();
		return;
	}
	definition = canonical<string>(cp);
	if (configured()) {
		/* -D option defining prior global  */
		if (defined()) { /* symbol is already defined */
			if (definition != *defn() ||
			    macro_params.length() != macro_parameters().length()) {
				/* Definitions differ */
				error_invalid_args() << "\"-D" << id() << macro_params.str()
					 << '=' << definition
					 << "\" contradicts prior \"-D" << id()
					 << macro_parameters().str()
					 << '=' << *defn() << '\"' << emit();
				return;
			} else { /* Definitions the same */
				warning_duplicate_args()
					<<  "Duplicated \"-D" << id() << macro_parameters().str()
					<< '=' << definition << "\" ignored" << emit();
				return;
			}
		} else  {
			/* -D contradicts prior -U option */
			error_invalid_args() << "\"-D" << id() << macro_params.str()
					 << '=' << definition << "\" contradicts prior \"-U"
					 << id() << '\"' << emit();
			return;
		}
	}
	define(definition);
	parameterize(macro_params);
	if (!configured()) {
		record(CONFIGURED_GLOBAL);
	}
}

line_type
symbol::digest_transient_define(parameter_list const & macro_params,
                                string const & definition)
{
	if (configured()) {
		/* This is a re-configuration of an already configured symbol */
		if (defined()) { /* symbol is already defined */
			if (definition != *defn() ||
			    macro_params.length() != macro_parameters().length()) {
				if (!global()) {
					/* Differing #define contradicts prior #define */
					warning_differing_redef() <<  "Differently redefining \""
						  << id() << macro_params.str()
						  << '=' << definition << "\", after \""
						  << id() << macro_parameters().str()
						  << '=' << *defn() << "\" at line "
						  << source_line() << emit();
				} else {
					/* Differing #define contradicts -D option */
					contradiction::flush();
					contradiction::insert(DIFFERENTLY_REDEFING_D,id());
					return LT_DIRECTIVE_DROP;
				}
			}
			/* Definitions the same */
			else if (!global()) {
				/* #define duplicating #define */
				return LT_DIRECTIVE_KEEP;
			} else { /* #define duplicating -D option */
				if (id() == contradiction::last_conflicted_symbol_id()) {
					contradiction::forget();
				}
				return LT_DIRECTIVE_DROP;
			}
		}
		/* symbol is already undefined. */
		else if (global()) {
			/* #define contradicts -U option */
			contradiction::insert(DEFINING_U,id());
			return LT_DIRECTIVE_DROP;
		}
		//* Else #define countervails #undef */
	}
	if (if_control::is_unconditional_line()) {
		if (options::no_transients()) {
			record(UNCONFIGURED);
		} else {
			if (!configured()) {
				warning_transient_symbol_added()
						<< "\"-D" << id() << macro_params.str()
						<< '=' << definition
						<< "\" has been assumed for the current file" << emit();
			}
			define(definition);
			parameterize(macro_params);
			record(CONFIGURED_TRANSIENT);
		}
	}
	return LT_DIRECTIVE_KEEP;
}

void symbol::digest_global_undef(char const *cp)
{
	if (*cp) {
		error_garbage_arg() << "Garbage in \"#undef " << id() << '\"' << emit();
		return;
	}
	if (configured()) {
		if (defined()) { /* symbol is already defined */
			error_invalid_args() << "\"-U" << id() << macro_parameters().str()
			                     << "\" contradicts prior \"-D"
			                     << id() << '=' << *defn() << '\"' << emit();
			return;
		}
		/* symbol is already undefined. */
		else {
			warning_duplicate_args() << "Duplicated \"-U"
			                         << id() << "\" ignored" << emit();
			return;
		}
	}
	undef();
	if (!configured()) {
		record(CONFIGURED_GLOBAL);
	}
}

line_type
symbol::digest_transient_undef()
{
	if (configured()) {
		if (defined()) { /* symbol is already defined */
			if (!global()) {
				/* #undef contradicting prior #define */
				warning_undefing_defined()
				        << "undefining " << id() << ", after defining "
				        << id() << macro_parameters().str()
				        << '=' << *defn() << " at line "
				        << source_line() << emit();
			} else  { /* #undef contradicting -D option */
				if (contradiction::last_conflicted_symbol_id() != id()) {
					contradiction::flush();
					contradiction::save(UNDEFING_D,id());
				}
				return LT_DIRECTIVE_DROP;
			}
		}
		/* symbol is already undefined. */
		else if (global()) {
			/* #undef echoes -U */
			return LT_DIRECTIVE_DROP;
		} else {
			/* #undef echoes #undef */
			return LT_DIRECTIVE_KEEP;
		}
	}
	if (if_control::is_unconditional_line()) {
		if (options::no_transients()) {
			record(UNCONFIGURED);
		} else {
			if (!configured()) {
				warning_transient_symbol_added() << "\"-U" << id() <<
						 "\" has been assumed for the current file" << emit();
			}
			undef();
			record(CONFIGURED_TRANSIENT);		
		}
	}
	return LT_DIRECTIVE_KEEP;
}

string symbol::macro_expand(argument_list const & macro_args) const
{
	if (!defn()) {
		error_macro_call_on_undefined()
		        << "Macro-call on undefined symbol " << id() << emit();
		return string();
	}
	if (macro_parameters().length() != macro_args.length()) {
		error_mismatched_macro_args() << "Symbol " << id() << " does not take "
					<< macro_args.length() << " macro parameters" << emit();
		return string();
	}
	return macro_args.subst(macro_parameters(),*defn());
}

void symbol::purge_transients()
{
	symbol_table::iterator iter = _sym_tab_.begin();
	symbol_table::iterator end = _sym_tab_.end();
	while(iter != end) {
		if (iter->second._provenance == CONFIGURED_TRANSIENT) {
			_sym_tab_.erase(iter++);
		} else {
			++iter;
		}
	}
}

size_t symbol::count(symbol_provenance source)
{
	symbol_table::const_iterator iter = _sym_tab_.begin();
	symbol_table::const_iterator end = _sym_tab_.end();
	size_t nsyms = 0;
	for (   ; iter != end; ++iter) {
		nsyms += iter->second._provenance == source;
	}
	return nsyms;
}

char const * symbol::scan_name(char const *cp)
{
	if (symbol::is_start_char(*cp)) {
		for (	; symbol::is_valid_char(*cp);
		        cp = chewer<continuation>(++cp)) {}
	}
	return cp;
}

void symbol::warn_variadic_unsupported()
{
	if (macro_parameters().variadic()) {
		warning_variadic_macros_unsupported() <<
			"Sorry, variadic macro expansion is unsupported. "
			<< "\"" << id() << macro_parameters().str() << "\"" <<
			" will not be resolved" << emit();
	}
}

bool symbol::add_pattern(string const & pattern) 
{
	return _selected_symbols_set_.insert(pattern).second;
}

bool symbol::remove_pattern(string const & pattern) 
{
	return _selected_symbols_set_.erase(pattern) == 1;
}

void symbol::set_selection(char const *optarg)
{
	string arg(optarg);
	size_t follower = 0;
	size_t leader = arg.find(',');
	for (	;leader != string::npos; 
			follower = leader + 1,leader = arg.find(',',follower)) {
		string pattern = arg.substr(follower,leader - follower);
		if (pattern.length()) {
			add_pattern(pattern);
		}
	}
	if (follower < arg.length()) {
		add_pattern(arg.substr(follower));
	}
}

bool symbol::wildcard_match(string const & wildcard, string const & name)
{
#if CXX11_HAVE_STRING_BACK
	if (wildcard.back() != '*') {
#else
	if (*(--wildcard.end()) != '*') {
#endif
		return false;
	}
	size_t min_len = wildcard.length() - 1;
	if (name.length() < min_len) {
		return false;
	}
	return memcmp(name.data(),wildcard.data(),min_len) == 0;
}
	

bool symbol::selected(string const & name)
{
	symbol_selection::const_iterator lb = 
		_selected_symbols_set_.lower_bound(name);
		
	if (lb != _selected_symbols_set_.end() && 
			(name == *lb || wildcard_match(*lb,name))) {
		return true;
	}
	return lb != _selected_symbols_set_.begin() && wildcard_match(*--lb,name);
}

bool symbol::deselected() const 
{
	return options::selected_symbols() && !selected(id());
} 

/* EOF*/
